from sklearn.preprocessing import StandardScaler


def standardize(fit_data, extra_data=None):
    """
    standardize data
    :param fit_data: data to fit and transform
    :param extra_data: extra data to transform
    :return:
    """
    s = StandardScaler()
    s.fit(fit_data)
    fit_data = s.transform(fit_data)
    if not extra_data is None:
        extra_data = s.transform(extra_data)
        return fit_data, extra_data
    return fit_data



import logging
import tensorflow as tf
import json
import numpy as np
from os.path import join
from model_zoo.utils import load_model

tfe = tf.contrib.eager


class ModelCheckpoint(tf.keras.callbacks.Callback):
    """
    Save model to checkpoints
    """
    
    def __init__(self,
                 checkpoint_dir,
                 checkpoint_name,
                 checkpoint_restore=True,
                 checkpoint_save_freq=2,
                 monitor='val_loss',
                 verbose=1,
                 checkpoint_save_best_only=False,
                 mode='auto',
                 period=1):
        """
        init checkpoint callback
        :param checkpoint_dir:
        :param checkpoint_name:
        :param checkpoint_restore:
        :param checkpoint_save_freq:
        :param monitor:
        :param verbose:
        :param checkpoint_save_best_only:
        :param mode:
        :param period:
        """
        super(ModelCheckpoint, self).__init__()
        self.monitor = monitor
        self.verbose = verbose
        self.checkpoint_dir = checkpoint_dir
        self.checkpoint_name = checkpoint_name
        self.file_path = join(checkpoint_dir, checkpoint_name)
        self.checkpoint_restore = checkpoint_restore
        self.checkpoint_save_freq = checkpoint_save_freq
        self.checkpoint_save_best_only = checkpoint_save_best_only
        self.period = period
        self.epochs_since_last_save = 0
        
        if mode not in ['auto', 'min', 'max']:
            logging.warning('ModelCheckpoint mode %s is unknown, '
                            'fallback to auto mode.', mode)
            mode = 'auto'
        
        if mode == 'min':
            self.monitor_op = np.less
            self.best = np.Inf
        elif mode == 'max':
            self.monitor_op = np.greater
            self.best = -np.Inf
        else:
            if 'acc' in self.monitor or self.monitor.startswith('fmeasure'):
                self.monitor_op = np.greater
                self.best = -np.Inf
            else:
                self.monitor_op = np.less
                self.best = np.Inf
    
    def on_train_begin(self, logs=None):
        """
        restore model from checkpoints
        :param logs:
        :return:
        """
        if self.checkpoint_restore:
            load_model(self.model, self.checkpoint_dir, self.checkpoint_name)
    
    def on_epoch_end(self, epoch, logs=None):
        """
        save model on epoch end
        :param epoch:
        :param logs:
        :return:
        """
        logs = logs or {}
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0
            file_path = self.file_path.format(epoch=epoch + 1, **logs)
            if self.checkpoint_save_best_only:
                current = logs.get(self.monitor)
                if current is None:
                    logging.warning('\nCan save best model only with %s available, '
                                    'skipping.', self.monitor)
                else:
                    if self.monitor_op(current, self.best):
                        if self.verbose > 0:
                            print('\nEpoch %05d: %s improved from %0.5f to %0.5f,'
                                  ' saving model to %s' % (epoch + 1, self.monitor, self.best,
                                                           current, file_path))
                        self.best = current
                        tfe.Saver(self.model.variables).save(file_path)
                        json.dump(dict(self.model.config),
                                  open('%s.json' % file_path, 'w', encoding='utf-8'),
                                  indent=2)
                    else:
                        if self.verbose > 0:
                            print('\nEpoch %05d: %s did not improve from %0.5f' %
                                  (epoch + 1, self.monitor, self.best))
            else:
                if self.verbose > 0:
                    print('\nEpoch %05d: saving model to %s' % (epoch + 1, file_path))
                tfe.Saver(self.model.variables).save(file_path)
                json.dump(dict(self.model.config),
                          open('%s.json' % file_path, 'w', encoding='utf-8'),
                          indent=2)
                if (epoch + 1) % self.checkpoint_save_freq == 0:
                    if self.verbose > 0:
                        print('Epoch %05d: saving model to %s-%d' % (epoch + 1, file_path, epoch + 1))
                    tfe.Saver(self.model.variables).save(file_path, global_step=epoch + 1)
                    json.dump(dict(self.model.config),
                              open('%s-%d.json' % (file_path, epoch + 1), 'w', encoding='utf-8'),
                              indent=2)



import tensorflow as tf
import model_zoo.callbacks as callbacks
import model_zoo.utils as utils
from tensorflow.python.framework import ops
from tensorflow.python.framework import tensor_util
from tensorflow.python.keras import backend as K
from tensorflow.python.keras.engine import base_layer
from tensorflow.python.keras.engine import training_utils
import math
import copy
import numpy as np

tfe = tf.contrib.eager

# override standardize_input_data method
training_utils.standardize_input_data = utils.standardize_input_data


class BaseModel(tf.keras.Model):
    """
    Base Keras Model, you can inherit this Model and
    override '__init__()', 'call()', 'init()', 'callback()' methods
    """
    
    def __init__(self, config):
        """
        init config, batch_size, epochs
        :param config:
        """
        super(BaseModel, self).__init__()
        self.config = config
        self.batch_size = config['batch_size']
        self.epochs = config['epochs']
        self.steps_per_epoch = config['steps_per_epoch']
        self.validation_steps = config['validation_steps']
        self.multiple_inputs = config['multiple_inputs']
    
    def construct(self, inputs):
        """
        Set inputs and output shape according to inputs
        :param inputs: inputs data or data piece
        :return:
        """
        if not self.multiple_inputs:
            if isinstance(inputs, (list, tuple)):
                if tensor_util.is_tensor(inputs[0]):
                    dummy_output_values = self.call(
                        training_utils.cast_if_floating_dtype(inputs[:1]))
                else:
                    dummy_output_values = self.call(
                        [ops.convert_to_tensor(v, dtype=K.floatx()) for v in inputs[:1]])
                # set dummy input values for inputs
                dummy_input_values = list(inputs[:1])
            else:
                if tensor_util.is_tensor(inputs):
                    dummy_output_values = self.call(
                        training_utils.cast_if_floating_dtype(inputs[:1]))
                else:
                    dummy_output_values = self.call(
                        ops.convert_to_tensor(inputs[:1], dtype=K.floatx()))
                # set dummy input values for inputs
                dummy_input_values = [inputs[:1]]
            # set output values
            if isinstance(dummy_output_values, (list, tuple)):
                dummy_output_values = list(dummy_output_values)
            else:
                dummy_output_values = [dummy_output_values]
        else:
            first_inputs = copy.copy(inputs)[0]
            if tensor_util.is_tensor(inputs):
                inputs = training_utils.cast_if_floating_dtype(inputs)
            else:
                inputs = ops.convert_to_tensor(inputs, dtype=K.floatx())
            inputs = tf.unstack(inputs[:, :1, :], axis=0)
            dummy_output_values = self.call(inputs, training=True)
            dummy_input_values = [first_inputs[:1]]
            # set output values
            if isinstance(dummy_output_values, (list, tuple)):
                dummy_output_values = list(dummy_output_values)
            else:
                dummy_output_values = [dummy_output_values]
        self.outputs = [
            base_layer.DeferredTensor(shape=(None for _ in v.shape),
                                      dtype=v.dtype) for v in dummy_output_values]
        self.inputs = [
            base_layer.DeferredTensor(shape=(None for _ in v.shape),
                                      dtype=v.dtype) for v in dummy_input_values]
        self.input_names = [
            'input_%d' % (i + 1) for i in range(len(dummy_input_values))]
        self.output_names = [
            'output_%d' % (i + 1) for i in range(len(dummy_output_values))]
        
        # self.call(tensor, training=True)
        self.built = True
        self.init()
    
    def call(self, inputs, training=None, mask=None):
        """
        build your models
        :param inputs: inputs x
        :param training: training flag
        :param mask: mask flag
        :return: y_pred
        """
        return inputs
    
    def init(self):
        """
        Default call compile method using sgd optimizer and mse loss.
        :return:
        """
        self.compile(optimizer=self.optimizer(), loss='mse')
    
    def train(self, train_data, eval_data=None, use_generator=False, **kwargs):
        """
        Train and fit model.
        :param train_data: x, y data pairs for training
        :param eval_data: x, y data pairs for evaluating
        :return: fit result
        """
        print('Training...')
        if not use_generator:
            # print('Train data', train_data)
            print(len(train_data[0][0]))
            x, y = train_data
            self.construct(x)
            # execute training
            return self.fit(x=x,
                            y=y,
                            epochs=self.epochs,
                            batch_size=self.batch_size,
                            validation_data=eval_data,
                            callbacks=self.callbacks())
        # use generator
        x, y = next(train_data)
        self.construct(x)
        
        # get train size, eval size
        train_size = kwargs.get('train_size', 0)
        eval_size = kwargs.get('eval_size', 0)
        
        # calculate steps_per_epoch
        steps_per_epoch = self.steps_per_epoch
        if not steps_per_epoch and train_size:
            steps_per_epoch = math.ceil(train_size / self.batch_size)
        if not steps_per_epoch:
            raise Exception('You must specify `steps_per_epoch` argument if `train_size` is not set')
        
        # calculate validation steps
        validation_steps = self.validation_steps
        if not validation_steps and eval_size:
            validation_steps = math.ceil(eval_size / self.batch_size)
        if not validation_steps:
            validation_steps = 1
        
        # fit generator
        return self.fit_generator(train_data,
                                  steps_per_epoch=steps_per_epoch,
                                  epochs=self.epochs,
                                  validation_data=eval_data,
                                  validation_steps=validation_steps,
                                  callbacks=self.callbacks()
                                  )
    
    def infer(self, test_data, batch_size=None):
        """
        do inference, default call predict method
        :param test_data: x data
        :param batch_size: batch_size
        :return:
        """
        x = test_data
        if not batch_size:
            batch_size = self.batch_size
        return self.predict(x, batch_size)
    
    def callbacks(self):
        """
        default callbacks, including logger, early stop, tensor board, checkpoint
        :return:
        """
        cbs = []
        cbs.append(tf.keras.callbacks.BaseLogger())
        if self.config.get('early_stop_enable'):
            cbs.append(tf.keras.callbacks.EarlyStopping(patience=self.config.get('early_stop_patience', 50)))
        if self.config.get('tensor_board_enable'):
            cbs.append(tf.keras.callbacks.TensorBoard(log_dir=self.config.get('tensor_board_dir', 'events')))
        if self.config.get('checkpoint_enable'):
            cbs.append(callbacks.ModelCheckpoint(
                checkpoint_dir=self.config.get('checkpoint_dir', 'checkpoints'),
                checkpoint_name=self.config.get('checkpoint_name', 'model'),
                checkpoint_restore=self.config.get('checkpoint_restore', True),
                checkpoint_save_freq=self.config.get('checkpoint_save_freq', 2),
                checkpoint_save_best_only=self.config.get('checkpoint_save_best_only', False)))
        return cbs
    
    def optimizer(self):
        """
        build optimizer, default to sgd and lr 0.01
        :return:
        """
        return tf.train.GradientDescentOptimizer(self.config.get('learning_rate', 0.01))



import tensorflow as tf
from model_zoo.utils import load_config, load_model, find_model

tfe = tf.contrib.eager
tf.enable_eager_execution()

tf.flags.DEFINE_string('checkpoint_dir', 'checkpoints', help='Data source dir', allow_override=True)
tf.flags.DEFINE_string('checkpoint_name', 'model.ckpt', help='Model name', allow_override=True)


class BaseEvaluater(object):
    """
    Base Evaluater, you need to specify
    """
    
    def __init__(self):
        """
        you need to define model_class in your Inferer
        """
        self.flags = tf.flags.FLAGS
    
    def prepare_data(self):
        """
        you need to implement this method
        :return:
        """
        raise NotImplementedError
    
    def run(self):
        """
        start inferring
        :return:
        """
        # prepare data
        self.eval_data = self.prepare_data()
        # split data
        x_eval, y_eval = self.eval_data
        # init configs from checkpoints json file and flags
        config = load_config(self.flags)
        # init model class
        model_class_name, model_file = config['model_class'], config['model_file']
        self.model_class = find_model(model_class_name, model_file)
        # init model
        model = self.model_class(config)
        # init variables
        model.construct(x_eval)
        # restore model
        load_model(model, self.flags.checkpoint_dir, self.flags.checkpoint_name)
        # evaluate
        return model.evaluate(x_eval, y_eval)



import json
from os.path import join
import numpy as np
import tensorflow as tf
from os.path import exists
from importlib import import_module
from tensorflow.python.framework import tensor_util
from tensorflow.python.keras.engine.training_utils import standardize_single_array

tfe = tf.contrib.eager


def load_config(FLAGS):
    """
    log configs
    :param FLAGS: FLAGS object
    :return: config dict
    """
    config = json.load(open('%s.json' % join(FLAGS.checkpoint_dir, FLAGS.checkpoint_name), 'r', encoding='utf-8'))
    for key, value in FLAGS.flag_values_dict().items():
        config[key] = value
    return config


def load_model(model, checkpoint_dir, checkpoint_name):
    """
    restore model from saved checkpoints
    :param model: model graph
    :return:
    """
    saver = tfe.Saver(model.variables)
    specified_checkpoint = join(checkpoint_dir, checkpoint_name)
    latest_checkpoint = tf.train.latest_checkpoint(checkpoint_dir)
    if specified_checkpoint and exists(checkpoint_dir) and tf.train.checkpoint_exists(specified_checkpoint):
        saver.restore(specified_checkpoint)
        print('Restored from %s' % specified_checkpoint)
    elif latest_checkpoint and exists(checkpoint_dir) and tf.train.checkpoint_exists(latest_checkpoint):
        saver.restore(latest_checkpoint)
        print('Restored from %s' % latest_checkpoint)
    else:
        print('No model to restore')


def find_model(model_class_name, model_file):
    """
    dynamic find model from model file according to model class name
    :param name: name of model class
    :param name: file name of model class
    :return: model class
    """
    if not model_class_name:
        raise Exception('You must define model_class in flags')
    model_file = import_module(model_file)
    model_class = getattr(model_file, model_class_name)
    return model_class


def get_shape(data):
    """
    get shape of data except batch_size
    :param data:
    :return:
    """
    if isinstance(data, np.ndarray) or isinstance(data, tf.Tensor):
        shape = data.shape
        return shape[1:]
    else:
        data = np.asarray(data)
        shape = data.shape
        return shape[1:]


def standardize_input_data(data,
                           names,
                           shapes=None,
                           check_batch_axis=True,
                           exception_prefix=''):
    """Normalizes inputs and targets provided by users.
  
    Users may pass data as a list of arrays, dictionary of arrays,
    or as a single array. We normalize this to an ordered list of
    arrays (same order as `names`), while checking that the provided
    arrays have shapes that match the network's expectations.
  
    Arguments:
        data: User-provided input data (polymorphic).
        names: List of expected array names.
        shapes: Optional list of expected array shapes.
        check_batch_axis: Boolean; whether to check that
            the batch axis of the arrays matches the expected
            value found in `shapes`.
        exception_prefix: String prefix used for exception formatting.
  
    Returns:
        List of standardized input arrays (one array per model input).
  
    Raises:
        ValueError: in case of improperly formatted user-provided data.
    """
    if not names:
        if (data is not None and hasattr(data, '__len__') and len(data) and
            not isinstance(data, dict)):
            raise ValueError('Error when checking model ' + exception_prefix + ': '
                                                                               'expected no data, but got:', data)
        return []
    if data is None:
        return [None for _ in range(len(names))]
    
    if isinstance(data, dict):
        try:
            data = [
                data[x].values
                if data[x].__class__.__name__ == 'DataFrame' else data[x]
                for x in names
            ]
        except KeyError as e:
            raise ValueError('No data provided for "' + e.args[0] + '". Need data '
                                                                    'for each key in: ' + str(names))
    elif isinstance(data, (list, tuple)):
        if isinstance(data[0], (list, tuple)):
            data = [np.asarray(d) for d in data]
        elif len(names) == 1 and isinstance(data[0], (float, int)):
            data = [np.asarray(data)]
        else:
            data = [
                x.values if x.__class__.__name__ == 'DataFrame' else x for x in data
            ]
    else:
        data = data.values if data.__class__.__name__ == 'DataFrame' else data
        data = [data]
    data = [standardize_single_array(x) for x in data]
    
    # if len(data) != len(names):
    #   if data and hasattr(data[0], 'shape'):
    #     raise ValueError('Error when checking model ' + exception_prefix +
    #                      ': the list of Numpy arrays that you are passing to '
    #                      'your model is not the size the model expected. '
    #                      'Expected to see ' + str(len(names)) + ' array(s), '
    #                      'but instead got the following list of ' +
    #                      str(len(data)) + ' arrays: ' + str(data)[:200] + '...')
    #   elif len(names) > 1:
    #     raise ValueError(
    #         'Error when checking model ' + exception_prefix +
    #         ': you are passing a list as input to your model, '
    #         'but the model expects a list of ' + str(len(names)) +
    #         ' Numpy arrays instead. The list you passed was: ' + str(data)[:200])
    #   elif len(data) == 1 and not hasattr(data[0], 'shape'):
    #     raise TypeError('Error when checking model ' + exception_prefix +
    #                     ': data should be a Numpy array, or list/dict of '
    #                     'Numpy arrays. Found: ' + str(data)[:200] + '...')
    #   elif len(names) == 1:
    #     data = [np.asarray(data)]
    
    # Check shapes compatibility.
    if shapes:
        for i in range(len(names)):
            if shapes[i] is not None:
                if tensor_util.is_tensor(data[i]):
                    tensorshape = data[i].get_shape()
                    if not tensorshape:
                        continue
                    data_shape = tuple(tensorshape.as_list())
                else:
                    data_shape = data[i].shape
                shape = shapes[i]
                if len(data_shape) != len(shape):
                    raise ValueError('Error when checking ' + exception_prefix +
                                     ': expected ' + names[i] + ' to have ' +
                                     str(len(shape)) + ' dimensions, but got array '
                                                       'with shape ' + str(data_shape))
                if not check_batch_axis:
                    data_shape = data_shape[1:]
                    shape = shape[1:]
                for dim, ref_dim in zip(data_shape, shape):
                    if ref_dim != dim and ref_dim is not None and dim is not None:
                        raise ValueError(
                            'Error when checking ' + exception_prefix + ': expected ' +
                            names[i] + ' to have shape ' + str(shape) +
                            ' but got array with shape ' + str(data_shape))
    return data



import tensorflow as tf
import types
from model_zoo.utils import find_model
import math

tfe = tf.contrib.eager
tf.enable_eager_execution()

# ========== Base Configs =================
tf.flags.DEFINE_integer('batch_size', 32, help='Batch size', allow_override=True)
tf.flags.DEFINE_float('learning_rate', 0.01, help='Learning rate', allow_override=True)
tf.flags.DEFINE_integer('epochs', 100, help='Max epochs', allow_override=True)
tf.flags.DEFINE_integer('validation_steps', 1, help='Validation steps', allow_override=True)
tf.flags.DEFINE_integer('steps_per_epoch', 0, help='Steps per epoch while using generator', allow_override=True)
tf.flags.DEFINE_bool('multiple_inputs', False, help='If inputs are multiple, set True, otherwise False', allow_override=True)

# ========== Early Stop Configs ================
tf.flags.DEFINE_bool('early_stop_enable', True, help='Whether to enable early stop', allow_override=True)
tf.flags.DEFINE_integer('early_stop_patience', 20, help='Early stop patience', allow_override=True)

# ========== Checkpoint Configs =================
tf.flags.DEFINE_bool('checkpoint_enable', True, help='Whether to save model checkpoint', allow_override=True)
tf.flags.DEFINE_string('checkpoint_dir', 'checkpoints', help='Data source dir', allow_override=True)
tf.flags.DEFINE_string('checkpoint_name', 'model.ckpt', help='Model name', allow_override=True)
tf.flags.DEFINE_bool('checkpoint_restore', False, help='Model restore', allow_override=True)
tf.flags.DEFINE_integer('checkpoint_save_freq', 2, help='Save model every epoch number', allow_override=True)

# ========== TensorBoard Events Configs =================
tf.flags.DEFINE_bool('tensor_board_enable', True, help='Whether to enable TensorBoard events', allow_override=True)
tf.flags.DEFINE_string('tensor_board_dir', 'events', help='TensorBoard events dir', allow_override=True)

# ========== Other Basic Configs ==================
tf.flags.DEFINE_string('model_file', 'model', help='Path of model file which including model class',
                       allow_override=True)
tf.flags.DEFINE_string('model_class', 'Model', help='Model class name, default to Model',
                       allow_override=True)

class BaseTrainer(object):
    """
    Base Trainer using eager mode, you can override tf.flags to define new config
    and you need to implement prepare_data method to prepare train_data and eval_data
    """
    
    def __init__(self):
        """
        you need to define model_class in your Trainer
        """
        self.flags = tf.flags.FLAGS
        # init model class
        model_class_name = self.flags.model_class
        self.model_class = find_model(model_class_name, self.flags.model_file)
    
    def prepare_data(self):
        """
        you need to implement this method
        :return:
        """
        raise NotImplementedError
    
    def build_generator(self, x_data, y_data, batch_size=None):
        """
        use this method to build a generator
        :param x_data:
        :param y_data:
        :param batch_size:
        :return:
        """
        batch_size = batch_size or self.flags.batch_size
        batches = math.ceil(len(x_data) / batch_size)
        while True:
            for i in range(int(batches)):
                yield x_data[i * batch_size: (i + 1) * batch_size], \
                      y_data[i * batch_size: (i + 1) * batch_size]
    
    def run(self):
        """
        this methods firstly init model class using flag values, then call model's train method to fit training data
        :return:
        """
        # prepare data
        data = self.prepare_data()
        # unpack data
        self.train_data, self.eval_data, self.train_size, self.eval_size = None, None, 0, 0
        if not isinstance(data, tuple):
            data = data,
        if len(data) == 1:
            self.train_data = data[0]
        elif len(data) == 2:
            self.train_data, self.eval_data = data
        elif len(data) == 3:
            self.train_data, self.eval_data, self.train_size = data
        elif len(data) == 4:
            self.train_data, self.eval_data, self.train_size, self.eval_size = data
        # build model and run
        model = self.model_class(self.flags.flag_values_dict())
        if isinstance(self.train_data, types.GeneratorType):
            model.train(self.train_data, self.eval_data, use_generator=True,
                        train_size=self.train_size, eval_size=self.eval_size)
        else:
            model.train(self.train_data, self.eval_data)
